package be.iminds.ilabt.jfed.highlevel.tasks;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.controller.JFedHighLevelException;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.model.*;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.highlevel.util.SliceRegistryUtil;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractFederationApi;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api.AggregateManager3;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.util.library.KeyUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 6/17/2015
 * Time: 11:50
 */
public final class ProvisionSliverTask extends CreateTask {
    private static final Logger LOG = LoggerFactory.getLogger(ProvisionSliverTask.class);
    //input parameters
    @Nonnull private final Slice slice;
    @Nonnull private final Server auth;
    @Nonnull private final Instant expirationTime;
    @Nonnull private final List<UserSpec> userSpecs;

    //results
    private String manifestRspec;

    private Set<Sliver> provisionedSlivers = null;

    @Nonnull private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    @Nonnull private final SfaModel sfaModel;
    @Nonnull private final HighLevelTaskFactory highLevelTaskFactory;
    @Nonnull private final SliceRegistryUtil sliceRegistryUtil;
    @Nonnull private final GeniUserProvider geniUserProvider;


    /**
     * Use this constructor to create a sliver that will be used immediately!
     */
    public ProvisionSliverTask(@Nonnull Slice slice,
                               @Nonnull Instant expirationTime,
                               @Nonnull Server auth,
                               @Nonnull List<UserSpec> userSpecs,
                               @Nonnull AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                               @Nonnull SfaModel sfaModel,
                               @Nonnull HighLevelTaskFactory highLevelTaskFactory,
                               @Nonnull SliceRegistryUtil sliceRegistryUtil,
                               @Nonnull GeniUserProvider geniUserProvider) {
        super("Provision Sliver @ " + auth.getDefaultComponentManagerUrn());
        this.slice = slice;
        this.expirationTime = expirationTime;
        this.auth = auth;
        this.userSpecs = generateUsedUserSpec(auth, userSpecs, geniUserProvider);

        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.sfaModel = sfaModel;
        this.highLevelTaskFactory = highLevelTaskFactory;
        this.sliceRegistryUtil = sliceRegistryUtil;
        this.geniUserProvider = geniUserProvider;
    }

    /**
     * This handles workarounds for servers that do not fully support the SSH accounts/keys in the Provision argument.
     *
     * @param auth
     * @param userSpecs
     * @param geniUserProvider
     * @return a modified list of userSpecs to send in the Provision request
     */
    private static List<UserSpec> generateUsedUserSpec(
            @Nonnull Server auth,
            @Nonnull List<UserSpec> userSpecs,
            GeniUserProvider geniUserProvider) {
        if (auth.hasFlag(Server.Flag.workaroundProvisionIgnoreSSHKey)) {
            LOG.info("OVERWRITING the Provision userSpecs list sent to "+auth.getName()+" because of flag workaroundProvisionIgnoreSSHKey");
            // use only the key found in the SSL certificate (since that is what the server will use anyway)
            GeniUser user = geniUserProvider.getLoggedInGeniUser();
            return UserSpecHelper.getUserUserSpecList(user);
        }
        if (auth.hasFlag(Server.Flag.workaroundProvisionSingleSSHKey)) {
            LOG.info("OVERWRITING the Provision userSpecs list sent to "+auth.getName()+" because of flag workaroundProvisionSingleSSHKey");
            //keep only the accounts and a single SSH key for the currently logged in user
            String userUrn = geniUserProvider.getLoggedInGeniUser().getUserUrnString();
            Optional<UserSpec> res = userSpecs.stream().filter(us -> Objects.equals(us.getUrn(), userUrn)).findFirst();
            if (res.isPresent()) {
                UserSpec orig = res.get();
                GeniUser user = geniUserProvider.getLoggedInGeniUser();
                String sslCertSshKey = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(user.getPublicKey());
                //prefer any key that is not the SSH key in the SSL cert, but use that if there is no other.
                String singleSshKey = orig.getSshKey().stream().filter(k -> !k.equals(sslCertSshKey)).findFirst().orElse(sslCertSshKey);
                return Collections.singletonList(new UserSpec(orig.getUrn(), singleSshKey));
            }
        }
        if (auth.hasFlag(Server.Flag.workaroundProvisionSingleSSHAccount)) {
            LOG.info("OVERWRITING the Provision userSpecs list sent to "+auth.getName()+" because of flag workaroundProvisionSingleSSHAccount");
            //keep only the accounts and SSH keys for the currently logged in user
            String userUrn = geniUserProvider.getLoggedInGeniUser().getUserUrnString();
            return userSpecs.stream().filter(us -> Objects.equals(us.getUrn(), userUrn)).collect(Collectors.toList());
        }
        //return the userSpecs list unmodified
        return new ArrayList<>(userSpecs);
    }

    @Override
    public void doTask(@Nonnull TaskExecution taskExecution) throws JFedException, InterruptedException {
        // Preparation: get a list of existing slivers on the given authority for the given slice.
        // This allows us to see which slivers were created during the Provision-call
        List<Sliver> existingSlivers = slice.findSlivers(auth).stream()
                //make a copy, otherwise, the sliver gets updated by the following calls
                .map(Sliver::createPartialCopy)
                .collect(Collectors.toList());

        AggregateManagerWrapper amWrapper =
                aggregateManagerWrapperFactory.getAggregateManagerWrapper(taskExecution, auth);

        assert amWrapper.hasSeperateAllocateAndProvision() : "AM Wrapper does not support provisioning!";


        LOG.info("Provision Sliver(s) @ {}  with user SSH keys: {}", auth.getDefaultComponentManagerUrn(), sfaModel.getUserKeys()); //missing here
        try {
            //don't set date here if CreateSliver doesn't set it itself
            Instant usedExpirationDate =
                    amWrapper.isExirationdateSetByCreateSliver()
                            || auth.hasFlag(Server.Flag.workaroundForceRenewAfterCreateSliver)
                            ? expirationTime : null;

            AggregateManagerWrapper.ProvisionReply provisionReply = amWrapper.provision(
                    slice.getUrn(),
                    slice.getCredentials(),
                    userSpecs,
                    usedExpirationDate != null ? Date.from(usedExpirationDate) : null
            );
            Instant provisionDate = Instant.now();
            this.manifestRspec = provisionReply.getManifest();

            sliceRegistryUtil.registerSliceAtSA(AbstractFederationApi.SliceRspecType.MANIFEST, slice, auth, manifestRspec, provisionDate, this);

            if (this.manifestRspec == null)
                throw new JFedHighLevelException("Provision of the sliver failed.");


            if (!isProvisioned(provisionReply.getSliverInfo())) {
                LOG.debug("Slivers are not yet Provisioned. Will wait before calling PerformOperationalAction geni_start until they are.");

                //initially, wait 1 seconds (in case of a fast provisioning AM), but wait 5 afterwards.
                Thread.sleep(1000l);

                AggregateManager3.StatusInfo sliverInfos =
                        amWrapper.status_v3(slice.getUrn(), slice.getCredentials());
                while (sliverInfos != null && !isProvisioned(sliverInfos.getSliverInfo())) {
                    LOG.debug("Slivers are not yet Provisioned. " +
                            "Will wait before calling PerformOperationalAction geni_start until they are.");

                    sliverInfos = amWrapper.status_v3(slice.getUrn(), slice.getCredentials());

                    Thread.sleep(5000l);
                }
            }

            //PerformOperationalAction geni_start
            amWrapper.poaStart(slice.getUrn(), slice.getCredentials());

            provisionedSlivers = new HashSet<>();

            Collection<Sliver> manifestSlivers =
                    SfaModelAbstractListener.findSliversInManifest(sfaModel, auth, slice.getUrn(), new ManifestRspecSource(manifestRspec, ModelRspecType.FX));
            if (manifestSlivers != null)
                provisionedSlivers.addAll(manifestSlivers);

            //add slivers found by sfaModel due to SliverStatus and Status calls.
            // AL2S is one case where this is needed, see FEDIBBTDEV-1392
            provisionedSlivers.addAll(slice.getSliversStream()
                    .filter(sliver -> Objects.equals(sliver.getServer(), auth))
                    .collect(Collectors.toSet()));

            List<Sliver> updatedSlivers = new ArrayList<>(provisionedSlivers);

            //we don't want to process slivers that already existed!
            provisionedSlivers.removeAll(existingSlivers);

            //we need to update slivers if they existed and their expiration date changed.
            updatedSlivers.removeIf(sliver -> {
                Optional<Sliver> orig = existingSlivers.stream().filter(s -> Objects.equals(s.getUrn(), sliver.getUrn())).findAny();
                LOG.debug("POSTProv SLIVER_INFO: updatedSlivers creation looked " +
                        "for "+sliver.getUrn()+" in existingSlivers and found "+orig);
                if (orig.isPresent()) {
                    //no need to update if expiration dates match
                    LOG.debug("POSTProv SLIVER_INFO: sliver.getExpirationDate()="+sliver.getExpirationDate()+" " +
                            "orig.get().getExpirationDate()="+orig.get().getExpirationDate());
                    return Objects.equals(orig.get().getExpirationDate(), sliver.getExpirationDate());
                } else {
                    //no matching sliver, so it will be in provisionedSlivers
                    return true;
                }
            });

            LOG.debug("POSTProv SLIVER_INFO: provisionedSlivers="+provisionedSlivers);
            LOG.debug("POSTProv SLIVER_INFO: existingSlivers="+existingSlivers);
            LOG.debug("POSTProv SLIVER_INFO: updatedSlivers="+updatedSlivers);

            if (!provisionedSlivers.isEmpty()) {
                sliceRegistryUtil.registerSliversAtSATaskExecution(provisionedSlivers, expirationTime, this);
                LOG.info("POSTProv SLIVER_INFO: Found new slivers during the Provision-call!");
            } else
                LOG.debug("POSTProv SLIVER_INFO: Could not find any slivers that were newly created in Provision call. This is OK, as they are typically created during the Allocate-call.");

            if (!updatedSlivers.isEmpty()) {
                sliceRegistryUtil.updateSliversAtSATaskExecution(updatedSlivers, expirationTime, this);
                LOG.info("POSTProv SLIVER_INFO: Found updated slivers during the Provision-call");
            } else
                LOG.debug("POSTProv SLIVER_INFO: Could not find any updated slivers in Provision call. This is OK.");
        } catch (AggregateManagerWrapper.NotEnoughFreeResourcesException e) {
            failedDueToNotEnoughFreeResources = true;
            this.manifestRspec = null;
            throw new JFedHighLevelException("Creation of the sliver failed: could not map to resources " +
                    "(resource shortage or unsupported topology)", e);
        } catch (AggregateManagerWrapper.TermsAndConditionsException e) {
            //typically this only occurs in Allocate, but not a bad thing to assume some AM's might also fail other calls.
            failedDueToTermAndConditionsNotAccepted = true;
            this.manifestRspec = null;
            throw new JFedHighLevelException("Creation of the sliver failed: Testbed access refused due to Terms & Conditions not accepted", e);
        } catch (AggregateManagerWrapper.ReservationProblemException e) {
            failedDueToReservationProblem = true;
            this.manifestRspec = null;
            throw new JFedHighLevelException("Creation of the sliver failed: Problem related to reservation of resources.", e);
        }
    }

    private boolean isProvisioned(@Nonnull List<AggregateManager3.SliverInfo> sliverInfos) {
        for (AggregateManager3.SliverInfo sliverInfo: sliverInfos) {
            if (sliverInfo.getAllocationStatus().equalsIgnoreCase("geni_allocated"))
                return false;
            if (sliverInfo.getOperationalStatus().equalsIgnoreCase("geni_pending_allocation"))
                return false;
        }
        return true;
    }

    @Nonnull
    @Override
    public List<Task> initDependsOn() {
        return Collections.singletonList(highLevelTaskFactory.getSliceCredential(slice));
    }

    @Nonnull public Slice getSlice() {
        return slice;
    }

    @Nonnull public Server getServer() {
        return auth;
    }

    @Override
    @Nonnull
    public Collection<Sliver> getCreatedSlivers() {
        return provisionedSlivers != null ? Collections.unmodifiableSet(provisionedSlivers) : Collections.emptySet();
    }


}
